/* *********************************************************************** *
 * project: org.matsim.*
 * WorldWriter.java
 *                                                                         *
 * *********************************************************************** *
 *                                                                         *
 * copyright       : (C) 2007 by the members listed in the COPYING,        *
 *                   LICENSE and WARRANTY file.                            *
 * email           : info at matsim dot org                                *
 *                                                                         *
 * *********************************************************************** *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *   See also COPYING, LICENSE and WARRANTY file                           *
 *                                                                         *
 * *********************************************************************** */

package org.matsim.world;

import java.io.IOException;
import java.util.Iterator;

import org.apache.log4j.Logger;
import org.matsim.core.gbl.Gbl;
import org.matsim.core.utils.io.IOUtils;
import org.matsim.core.utils.io.Writer;

public class WorldWriter extends Writer {

	//////////////////////////////////////////////////////////////////////
	// member variables
	//////////////////////////////////////////////////////////////////////

	private WorldWriterHandler writerhandler = null;
	private final World world;

	private final static Logger log = Logger.getLogger(WorldWriter.class);

	//////////////////////////////////////////////////////////////////////
	// constructors
	//////////////////////////////////////////////////////////////////////

	/**
	 * Creates a new WorldWriter with the output-file-location from the current configuration.
	 *
	 * @param world The world to be written to file.
	 */
	public WorldWriter(final World world) {
		this(world, Gbl.getConfig().world().getOutputFile());
	}

	/**
	 * Creates a new WorldWriter to write the specified world to the specified file.
	 *
	 * @param world
	 * @param filename
	 */
	public WorldWriter(final World world, final String filename) {
		super();
		this.world = world;
		this.outfile = filename;
		// always write out latest version, currently v2
		this.dtd = "http://www.matsim.org/files/dtd/world_v2.dtd";
		this.writerhandler = new WorldWriterHandlerImplV2();
	}

	//////////////////////////////////////////////////////////////////////
	// private methods
	//////////////////////////////////////////////////////////////////////

	private final void writeLayer(final Layer l) {
		if (l instanceof ZoneLayer) {
			try {
				this.writerhandler.startLayer((ZoneLayer)l, this.out);
				Iterator<? extends Location> z_it = l.getLocations().values().iterator();
				while (z_it.hasNext()) {
					Zone z = (Zone)z_it.next();
					this.writerhandler.startZone(z, this.out);
					this.writerhandler.endZone(this.out);
				}
				this.writerhandler.endLayer(this.out);
				this.writerhandler.writeSeparator(this.out);
				this.out.flush();
			}
			catch (IOException e) {
				Gbl.errorMsg(e);
			}
		}
		else {
			log.info("layer_type=" + l.getType() + ": Layer must be explicitly written to a seperate xml file");
		}
	}

	private final void writeRule(final MappingRule m) {
		if ((m.getDownLayer() instanceof ZoneLayer)) {
			try {
				this.writerhandler.startMapping(m, this.out);
				Iterator<? extends Location> dz_it = m.getDownLayer().getLocations().values().iterator();
				while (dz_it.hasNext()) {
					Zone dz = (Zone)dz_it.next();
					Iterator<Location> uz_it = dz.getUpMapping().values().iterator();
					while (uz_it.hasNext()) {
						Zone uz = (Zone)uz_it.next();
						this.writerhandler.startRef(dz, uz, this.out);
						this.writerhandler.endRef(this.out);
						this.out.flush();
					}
				}
				this.writerhandler.endMapping(this.out);
				this.writerhandler.writeSeparator(this.out);
				this.out.flush();
			}
			catch (IOException e) {
				Gbl.errorMsg(e);
			}
		}
		else {
			log.info("[m=" + m + "," + "downLayer_type=" + m.getDownLayer().getType() + ": Layer not written. Therefore, rule is not written]");
		}
	}

	//////////////////////////////////////////////////////////////////////
	// write methods
	//////////////////////////////////////////////////////////////////////

	@Override
	public final void write() {
		try {
			this.out = IOUtils.getBufferedWriter(this.outfile);
			writeDtdHeader("world");
			this.out.flush();
			this.writerhandler.startWorld(this.world, this.out);
			this.writerhandler.writeSeparator(this.out);
			this.out.flush();
			Layer l = this.world.getBottomLayer();
			if (l != null) {
				this.writeLayer(l);
				while (l.getUpRule() != null) {
					l = l.getUpRule().getUpLayer();
					this.writeLayer(l);
				}
			}
			l = this.world.getBottomLayer();
			if (l != null) {
				if (l.getUpRule() != null) {
					MappingRule m = l.getUpRule();
					this.writeRule(m);
					while (m.getUpLayer().getUpRule() != null) {
						m = m.getUpLayer().getUpRule();
						this.writeRule(m);
					}
				}
			}
			this.writerhandler.endWorld(this.out);
			this.out.flush();
			this.out.close();
		}
		catch (IOException e) {
			Gbl.errorMsg(e);
		}
	}

	//////////////////////////////////////////////////////////////////////
	// print methods
	//////////////////////////////////////////////////////////////////////

	@Override
	public final String toString() {
		return super.toString();
	}
}
